// The MIT License

// Copyright (c) 2009 Massachusetts Institute of Technology

// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:

// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.

// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

import Connectable::*;
import GetPut::*;
import ClientServer::*;
import FIFO::*;
import FIFOF::*;
import SpecialFIFOs::*;

//AWB includes
`include "asim/provides/low_level_platform_interface.bsh"
`include "asim/provides/soft_connections.bsh"
`include "asim/provides/common_services.bsh"


// Local includes
`include "asim/provides/core.bsh"
`include "asim/provides/processor_library.bsh"
`include "asim/provides/processor_library.bsh"
`include "asim/provides/fpga_components.bsh"
`include "asim/rrr/remote_client_stub_PROCESSORSYSTEMRRR.bsh"
`include "asim/rrr/remote_server_stub_PROCESSORSYSTEMRRR.bsh"

typedef enum {Decompr1, Decompr2, Decompr3, Decompr4} DecomprStage deriving(Eq, Bits);
typedef enum {Resp1, Resp2, Resp3, Resp4} ComprStage deriving(Eq, Bits);

module [CONNECTED_MODULE] mkConnectedApplication ();

  Core core <- mkCore;
  Reg#(int) cycle <- mkReg(0);
  
//	// wjun
	ServerStub_PROCESSORSYSTEMRRR server_stub <- mkServerStub_PROCESSORSYSTEMRRR(); 

  ClientStub_PROCESSORSYSTEMRRR client_stub <- mkClientStub_PROCESSORSYSTEMRRR();   
  // Make this big enough so that several outstanding requests may be supported
  FIFO#(Bit#(MainMemTagSz)) itags <- mkSizedFIFO(8);

	FIFOF#(Decompressed) rescacheQ <- mkSizedFIFOF(16);
/*
	rule cache_result;
		let resp <- core.decompress_result.get;
		rescacheQ.enq(resp);
//		client_stub.makeRequest_SendDecompressed(resp.value0, resp.value1, resp.value2); 
	endrule
	rule send_result;
		rescacheQ.deq();
		let resp = rescacheQ.first();
		client_stub.makeRequest_SendDecompressed(resp.value0, resp.value1, resp.value2); 
	endrule
	rule result_resp_dummy;
		let compressedResp <- client_stub.getResponse_SendDecompressed();
	endrule
*/
	Reg#(DecomprStage) decomprstage <- mkReg(Decompr1);
	Reg#(StepRes) decomprcache1 <- mkReg(?);
	Reg#(StepRes) decomprcache2 <- mkReg(?);
	Reg#(StepRes) decomprcache3 <- mkReg(?);
	rule cacheDecompr1(decomprstage == Decompr1);
		decomprstage <= Decompr2;
		let res <- core.get_decompressed.get;
		decomprcache1 <= res;
		$display("decompr cache");
	endrule
	
	rule cacheDecompr2(decomprstage == Decompr2);
		decomprstage <= Decompr3;
		let res <- core.get_decompressed.get;
		decomprcache2 <= res;
		$display("decompr cache");
	endrule
	rule cacheDecompr3(decomprstage == Decompr3);
		decomprstage <= Decompr4;
		let res <- core.get_decompressed.get;
		decomprcache3 <= res;
		$display("decompr cache");
	endrule

	rule sendDecompr(decomprstage == Decompr4);
		let res <- core.get_decompressed.get;
		let res1 = decomprcache1;
		let res2 = decomprcache2;
		let res3 = decomprcache3;
		client_stub.makeRequest_SendResult(
			{pack(res1.done), 2'b00, res1.target, res1.size, res1.memAcc, pack(res1.regChange), res1.matchcount },
			res1.addr,
			res1.value,
			res1.pc,
			{pack(res2.done), 2'b00, res2.target, res2.size, res2.memAcc, pack(res2.regChange), res2.matchcount },
			res2.addr,
			res2.value,
			res2.pc,
			{pack(res1.done), 2'b00, res3.target, res3.size, res3.memAcc, pack(res3.regChange), res3.matchcount },
			res3.addr,
			res3.value,
			res3.pc,
			{pack(res.done), 2'b00, res.target, res.size, res.memAcc, pack(res.regChange), res.matchcount },
			res.addr,
			res.value,
			res.pc
		);
		$display("decompr send");
		decomprstage <= Decompr1;
	endrule
	/*
	rule sendDecompr_resp;
		let resp <- client_stub.getResponse_SendResult();
	endrule
  */
	// this is for the tracing
  rule printCycles;
    cycle <= cycle+1;
    $fdisplay(stderr, " => Cycle = %d", cycle);
  endrule

	Reg#(Bit#(16)) flying <- mkReg(0);

	Reg#(ComprStage) comprstage <- mkReg(Resp1);
	rule reqCompressed( flying < core.get_compressedQ_count() );
		$display ( "Compr Request" );
		client_stub.makeRequest_ReqCompressed(core.get_proc_data()); 
		
		flying <= flying + 4;
//		comprstage <= Resp1;
	endrule
/*
	rule sendQAvail( core.get_compressedQ_count() > 256 && flying == 0 );
//		client_stub.makeRequest_SendQAvail({16'h0, core.get_compressedQ_count()});
//		flying <= core.get_compressedQ_count();

		client_stub.makeRequest_SendQAvail(256);
		flying <= 256;
	endrule
	rule recvQAvail;
		let res <- client_stub.getResponse_SendQAvail;
	endrule
*/
	Reg#(StepRes) resp2 <- mkReg(?);
	Reg#(StepRes) resp3 <- mkReg(?);
	Reg#(StepRes) resp4 <- mkReg(?);
	rule respCompressed (comprstage == Resp1);
		let resp <- client_stub.getResponse_ReqCompressed();
		//let resp <- server_stub.acceptRequest_RespCompressed();
		$display ( "1 : %x-%x-%x-%x", resp.flags1, resp.addr1, resp.value1,resp.pc1);
		core.add_compressedQ_count();
		core.put_compressed.put(
			StepRes{
				matchcount:resp.flags1[15:0],
				regChange:(resp.flags1[16] == 1),
				memAcc:resp.flags1[18:17],
				size:resp.flags1[23:19],
				target:resp.flags1[28:24],
				done:(resp.flags1[31] == 1),

				pc:resp.pc1,
				addr:resp.addr1,
				value:resp.value1
			}
		);
		resp2 <= StepRes{
				matchcount:resp.flags2[15:0],
				regChange:(resp.flags2[16] == 1),
				memAcc:resp.flags2[18:17],
				size:resp.flags2[23:19],
				target:resp.flags2[28:24],
				done:(resp.flags2[31] == 1),

				pc:resp.pc2,
				addr:resp.addr2,
				value:resp.value2
		};
		resp3 <= StepRes{
				matchcount:resp.flags3[15:0],
				regChange:(resp.flags3[16] == 1),
				memAcc:resp.flags3[18:17],
				size:resp.flags3[23:19],
				target:resp.flags3[28:24],
				done:(resp.flags3[31] == 1),

				pc:resp.pc3,
				addr:resp.addr3,
				value:resp.value3
		};
		resp4 <= StepRes{
				matchcount:resp.flags4[15:0],
				regChange:(resp.flags4[16] == 1),
				memAcc:resp.flags4[18:17],
				size:resp.flags4[23:19],
				target:resp.flags4[28:24],
				done:(resp.flags4[31] == 1),

				pc:resp.pc4,
				addr:resp.addr4,
				value:resp.value4
		};
		flying <= flying - 1;
		comprstage <= Resp2;
	endrule
	rule resp2Queue2( comprstage == Resp2 );
		core.add_compressedQ_count();
//		$display ( "2 : %x-%x-%x", resp2.addr, resp2.value,resp2.pc);
		core.put_compressed.put(resp2);
		comprstage <= Resp3;
		flying <= flying - 1;
	endrule
	rule resp2Queue3( comprstage == Resp3 );
		core.add_compressedQ_count();
//		$display ( "2 : %x-%x-%x", resp2.addr, resp2.value,resp2.pc);
		core.put_compressed.put(resp3);
		comprstage <= Resp4;
		flying <= flying - 1;
	endrule
	rule resp2Queue4( comprstage == Resp4 );
		core.add_compressedQ_count();
		core.put_compressed.put(resp4);
		comprstage <= Resp1;
		flying <= flying - 1;
	endrule

//TODO::: remove 'tag'. Not needed
  rule sendInstReq;
  //rule sendInstReq (core.ispredicting());
    let coreReq <- core.imem_client.request.get;
//		$display( "Sending memReq" );
	
    case (coreReq) matches 
      tagged LoadReq .load: begin
                              //Allocate ROB space
                              client_stub.makeRequest_InstrRequestLoad(load.addr);
//                              itags.enq(load.tag);
                            end
	endcase
  endrule

	Reg#(Bit#(3)) instrFetchIdx <- mkReg(0);
	Reg#(Bit#(32)) v1 <- mkReg(0);
	Reg#(Bit#(32)) v2 <- mkReg(0);
	Reg#(Bit#(32)) v3 <- mkReg(0);
	Reg#(Bit#(32)) v4 <- mkReg(0);
	Reg#(Bit#(32)) v5 <- mkReg(0);
	Reg#(Bit#(32)) v6 <- mkReg(0);
	Reg#(Bit#(32)) v7 <- mkReg(0);
  rule receiveInstResp(instrFetchIdx == 0);
    let resp <- client_stub.getResponse_InstrRequestLoad();
		v1 <= resp.value1;
		v2 <= resp.value2;
		v3 <= resp.value3;
		v4 <= resp.value4;
		v5 <= resp.value5;
		v6 <= resp.value6;
		v7 <= resp.value7;
		instrFetchIdx <= 1;
//    itags.deq;
    core.imem_client.response.put(tagged LoadResp {data:resp.value0,
                                                   tag: 0});
  endrule
	rule instrFetchFeed(instrFetchIdx > 0);
		if ( instrFetchIdx < 7 )
			instrFetchIdx <= instrFetchIdx + 1;

		if ( instrFetchIdx == 1 ) 
    	core.imem_client.response.put(tagged LoadResp {data:v1, tag: 1});
		else if ( instrFetchIdx == 2 ) 
    	core.imem_client.response.put(tagged LoadResp {data:v2, tag: 2});
		else if ( instrFetchIdx == 3 ) 
    	core.imem_client.response.put(tagged LoadResp {data:v3, tag: 3});
		else if ( instrFetchIdx == 4 ) 
    	core.imem_client.response.put(tagged LoadResp {data:v4, tag: 4});
		else if ( instrFetchIdx == 5 ) 
    	core.imem_client.response.put(tagged LoadResp {data:v5, tag: 5});
		else if ( instrFetchIdx == 6 ) 
    	core.imem_client.response.put(tagged LoadResp {data:v6, tag: 6});
		else if ( instrFetchIdx == 7 ) begin
    	core.imem_client.response.put(tagged LoadResp {data:v7, tag: 7});
			instrFetchIdx <= 0;
		end
	endrule
endmodule
